MultiVarRegime <- function(data, EOFS = TRUE, neofs = 30, lon = NULL, lat = NULL, ncenters = NULL, method = "kmeans",
                          nstart = 30, multivariate = FALSE, listindex = NULL) {
  if (EOFS != 1) {
    stop("Currently clusters must be calculated from fields")
  }
  if (multivariate == FALSE) {
    sdate <- which(names(dim(data)) == "sdate")
    ftime <- which(names(dim(data)) == "ftime")
    nftimes <- dim(data)[ftime]
    nsdates <- dim(data)[sdate]
    lon2 <- which(names(dim(data)) == "lon")
    lat2 <- which(names(dim(data)) == "lat")
    dims <- c(1 : length(dim(data)))[-c(sdate, ftime, lon2, lat2)]
    data <- aperm(data, c(sdate, ftime, lat2, lon2, dims))
    nlon <- dim(data)[lon2]
    nlat <- dim(data)[lat2]
    data <- melt(data, varnames = c("Year", "Day", "Lat", "Lon"))
    data <- acast(data, Year + Day ~ Lat ~ Lon)
    if (is.null(ncenters)) {
      stop("ncenters must be specified")
    }
  } else {
    sdate <- which(names(dim(data)) == "sdate")
    ftime <- which(names(dim(data)) == "ftime")
    nftimes <- dim(data)[ftime]
    nsdates <- dim(data)[sdate]
    lon2 <- which(names(dim(data)) == "lon")
    lat2 <- which(names(dim(data)) == "lat")
    dataset <- which(names(dim(data)) == "var")
    nvars <- dim(data)[dataset]
    nlon <- dim(data)[lon2]
    nlat <- dim(data)[lat2] * nvars
    lat = rep(lat, nvars)
    data <- aperm(data, c(sdate, ftime, lat2, dataset, lon2))
    var_mean <- var_sd <- c()
    for (i in 1 : nvars) {
      var_mean[i] <- mean(data[ , , , i, ], na.rm = TRUE)
      var_sd[i] <- sd(data[, , , i, ], na.rm = TRUE)
      data[, , , i, ] <- (data[, , , i, ] - var_mean[i]) / var_sd[i]
    }
    data <- melt(data, varnames = c("Year", "Day", "Lat", "Dataset", "Lon"))
    data <- acast(data, Year + Day ~ Dataset + Lat ~ Lon)
  }
  if (EOFS  == TRUE) {
    #data <- princomp(data[,subs], cor = FALSE)$scores
    dataPC <- EOF(data, lat = lat, lon = lon, neofs = neofs)
    threshold <- 90
    minPC <- head(as.numeric(which(cumsum(dataPC$var) > threshold)), 1)
  } 
  if (method == "kmeans") {
    result <- kmeans(dataPC$PC[, 1 : minPC], centers = ncenters, 
                     iter.max = 100, nstart = nstart, trace = FALSE)
    if (multivariate == FALSE) {
      reconstructed <- array(0, c(ncenters, nlat, nlon))
      for (i in 1 : ncenters) {
        field <- apply(dataPC$EOF[1 : minPC, , ], MARGIN = c(2,3), function(x) x * result$center[i,])
        reconstructed[i, , ] <- apply(field, MARGIN = c(2,3), sum)
      }
      reconstructed <- aperm(reconstructed, c(1,3,2))
    } else {
      reconstructed <- array(0, c(ncenters, nlat, nlon))
      for (i in 1 : ncenters) {
        field <- apply(dataPC$EOF[1 : minPC, , ], MARGIN = c(2,3), function(x) x * result$center[i,])
        reconstructed[i, , ] <- apply(field, MARGIN = c(2,3), sum)
      }
      reconstructed2 <- array(0, c(nvars, ncenters, (nlat / nvars), nlon))
      for (i in 1 : nvars) {
        for (j in 1 : ncenters) {
          reconstructed2[i,j , , ] <- reconstructed[j, (((i - 1) * nlats) + 1) : (i * nlats), ] 
        }
        reconstructed2[i, , , ] <- (reconstructed2[i, , , ] * var_sd[i]) + var_mean[i]
      }
      reconstructed2 <- aperm(reconstructed2, c(1,2,4,3))
      reconstructed <- reconstructed2
      rm(reconstructed2)
    }
  
    names(dim(reconstructed)) <- c("cluster", "lon", "lat")
    cluster_timeseries <- list(lengths = c(), values = c())
    for (i in 1 : nsdates) {
      occurences <- rle(result$cluster[((i * nftimes) + 1 - nftimes) : (i * nftimes)])
      cluster_timeseries <- list(lengths = c(cluster_timeseries$lengths, occurences$lengths),
                                 values = c(cluster_timeseries$values, occurences$values))
    }
    frequency <- persistence <- list()
    for (i in 1 : ncenters) {
     frequency[[i]] <- (sum(cluster_timeseries$lengths[cluster_timeseries$values == i]) / length(result$cluster)) * 100
     persistence[[i]] <- mean(cluster_timeseries$length[cluster_timeseries$values == i])
    }
    
  } else if (method == "medoids") {
    result <- pam(dataPC$PC[, 1 : minPC], k = ncenters)
    reconstructed <- array(0, c(ncenters, nlat, nlon))
    for (i in 1 : ncenters) {
      field <- apply(dataPC$EOF[1 : minPC, , ], MARGIN = c(2,3), function(x) x * result$medoids[i,])
      reconstructed[i, , ] <- apply(field, MARGIN = c(2,3), sum)
    }
    reconstructed <- aperm(reconstructed, c(1,3,2))
    names(dim(reconstructed)) <- c("cluster", "lon", "lat")
    cluster_timeseries <- list(lengths = c(), values = c())
    for (i in 1 : nsdates) {
      occurences <- rle(result$clustering[((i * nftimes) + 1 - nftimes) : (i * nftimes)])
      cluster_timeseries <- list(lengths = c(cluster_timeseries$lengths, occurences$lengths),
                                 values = c(cluster_timeseries$values, occurences$values))
    }
    frequency <- persistence <- list()
    for (i in 1 : ncenters) {
      frequency[[i]] <- (sum(cluster_timeseries$lengths[cluster_timeseries$values == i]) / length(result$clustering)) * 100
      persistence[[i]] <- mean(cluster_timeseries$length[cluster_timeseries$values == i])
    }
  } else {
    result <- hclust(dist(dataPC$PC[, 1 : minPC]), method = method)
    clusterCut <- cutree(result, ncenters)
    data <- aperm(data, c(3, 2, 1))
    result <- Composite(data, clusterCut)
  }
  if (method == "kmeans") {
    return(list(field = reconstructed, EOF = dataPC$EOF, center = result$center, cluster = result$cluster,
           cluster_lengths = cluster_timeseries$lengths, cluster_values = cluster_timeseries$values,
           persistence = persistence, frequency = frequency))
  } else if (method == "medoids") {
    return(list(field = reconstructed, EOF = dataPC$EOF, center = result$medoids, cluster = result$clustering,
                cluster_lengths = cluster_timeseries$lengths, cluster_values = cluster_timeseries$values,
                persistence = persistence, frequency = frequency))
  } else {
    return(list(field = result, cluster = clusterCut))
  }
}


